from abc import ABC, abstractmethod, abstractproperty
import typing
import os
import io


# Signatures cause circular type dependencies
class Storage(ABC):
    pass

class StorageItem(ABC):
    pass

class StorageSearchIter(ABC):
    pass


class StorageItem(ABC):

    @staticmethod
    @abstractmethod
    def from_file(path, remote_path="", name=None):
        pass

    @staticmethod
    @abstractmethod
    def from_storage(o: Storage, path=""):
        pass


    @abstractmethod
    def __init__(self):
        super().__init__()

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    @abstractmethod
    def size(self):
        pass

    @property
    @abstractmethod
    def remote_path(self):
        pass

    def __str__(self):
        return self.name


class StorageSearchIter(ABC):
    """
    Represents a search of storage items. This iterator should not return the same individual item twice
    (that is, if there are two otherwise identical copies of an item on remote storage, each should be returned once)
    """

    @abstractmethod
    def __init__(self):
        super().__init__()

    def __iter__(self):
        return self

    @abstractmethod
    def __next__(self):
        """
        Iterate through whatever remote file tree is involved and return individual file objects
        :return: StorageObject representing a single file object
        :rtype: StorageObject
        """
        pass


class Storage(ABC):
    """
    Represents a "connection" to remote storage. All methods may implement additional OPTIONAL parameters, but generally
    should store within itself any other information needed to execute searches, puts, gets, and deletes.
    The exception being the __init__ method, which may involve more strict and platform-specific requirements, such as
    credentials.
    ...
    Attributes
    ----------
    ItemTypes
    """
    ItemTypes = [StorageItem]
    stream_support = False

    def __init__(self):
        super().__init__()

    @abstractmethod
    def search(self, basepath: str = ""):
        """
        Return an interator representing all objects within a specified search of the remote storage.
        Should generally return objects that are children of, or begin with basepath.
        :param basepath: Allow limiting to sub "directories" or other organizational schemas
        :return: StorageObjectIter representing all files within basepath
        :rtype: StorageObjectIter
        """

    @abstractmethod
    def put(self, o: StorageItem, filepath: str, prefix: str = ""):
        """
        Upload file from filepath to remote storage
        :param o: Remote storage representation
        :param filepath: local file location
        :param prefix: Prefix to append to remote file location
        :return: Success?
        :rtype: bool
        """
        pass

    def put_stream(self, o: StorageItem, file: io.IOBase, prefix: str = ""):
        """
        Download file from remote storage to local IO Stream
        :param o: Remote storage object representation
        :param file: IO object to write a binary stream to
        :param prefix: Prefix to append to remote file location
        :return: Success?
        :rtype: bool
        """
        raise NotImplemented

    @abstractmethod
    def get(self, o: StorageItem, filepath: str = ""):
        """
        Download file from remote storage to filepath
        :param o: Remote storage object representation
        :param filepath: optional information to associate with the remote object, eg s3 tags, expiry, or storage settings
        :return: Success?
        :rtype: bool
        """
        pass

    @abstractmethod
    def fetch(self, o: StorageItem):
        """
        Fetch details about a remote object and return a object containing that information
        The intent is for a user to create a base StorageItem object containing enough information to locate the file
            in question (say, a remote path for example), and then use this method to fetch all actual metadata about
            the object that was not known prior. The method should not modify o, but should create a new StorageItem
        :param o: Remote storage object representation
        :return: o?
        :rtype: StorageObject
        """
        pass

    def get_stream(self, o: StorageItem, file: io.IOBase):
        """
        Download file from remote storage to local IO Stream
        :param o: Remote storage object representation
        :param file: IO object to write a binary stream to
        :return: Success?
        :rtype: bool
        """
        raise NotImplemented

    @abstractmethod
    def delete(self, o: StorageItem):
        """
        Delete file from remote storage
        :param o: Remote storage representation
        :return: Success?
        :rtype: bool
        """
        pass


from .storage import StorageItem

# Generic classes covering some common cloud storage scenarios

class ObjectNotFoundError(Exception):
    pass


class ObjectPermissionError(Exception):
    pass


class UnsupportedItemTypeError(Exception):
    pass


import uuid
from .core.storage import Storage, StorageItem, StorageSearchIter
import os
import tempfile

# Copy from remote source to local disk, copy from local disk to remote destination, delete local disk copy
def _using_local_ehpemeral(source: Storage, objects: [list,StorageSearchIter], destination: Storage, delete_source=False, destination_prefix = ""):
    with tempfile.TemporaryDirectory(suffix='sscstu' + str(uuid.uuid4())) as tmpdir:
        for o in objects:
            print(f"Transfer {o} from {source} to {destination}")
            tmppath = tmpdir + str(uuid.uuid4())
            source.get(o, tmppath)
            if type(o) in destination.ItemTypes:
                if not destination.put(o, tmppath, prefix=destination_prefix):
                    raise Exception(f"Failed to put {o} from {tmppath}")
            else:
                # Use from_file on most preferred
                if not destination.put(type(destination.ItemTypes[0]).from_file(tmppath, name=o.name), tmppath, prefix=destination_prefix):
                    os.remove(tmppath)
                    raise Exception(f"Failed to put {o} from {tmppath}")
            os.remove(tmppath)
            if delete_source:
                if not source.delete(o):
                    raise Exception(f"Failed to delete {o} from source")

def _using_mem_ephemeral():
    raise NotImplementedError()

def _using_async_fstreams():
    raise NotImplementedError()

def transfer(source: [Storage, str], destination: [Storage, str], source_basepath=None, destination_prefix=""):
    if source_basepath is None:
        raise ValueError("Must specify a base search path for source Storage object")
    return _using_local_ehpemeral(source, source.search(source_basepath), destination, destination_prefix=destination_prefix)


from ..core.storage import Storage, StorageSearchIter, StorageItem
from pathlib import Path
import shutil
from shutil import SameFileError
import os
import io

class LocalStorage(Storage):
    pass

class LocalStorageItem(StorageItem):
    pass

class LocalStorageSearchIter(StorageSearchIter):
    pass


class LocalStorageItem(StorageItem):
    path = None
    @staticmethod
    def from_file(path, remote_path="", name=None):
        if remote_path == "":
            remote_path = path
        if name is None:
            name = remote_path
        return LocalStorageItem(name)


    @staticmethod
    def from_storage(o: Storage, path=""):
        pass

    def __init__(self, path=""):
        self.path = path
        super().__init__()

    @property
    def name(self):
        return Path(self.path).name

    @property
    def size(self):
        return os.stat(self.path).st_size

    @property
    def remote_path(self):
        return self.path

    def __str__(self):
        return self.name


class LocalStorageSearchIter(LocalStorage):
    """
    Represents a search of storage items. This iterator should not return the same individual item twice
    (that is, if there are two otherwise identical copies of an item on remote storage, each should be returned once)
    """
    ItemTypes = [LocalStorageItem]

    def __init__(self, basepath: str):
        l = os.listdir(basepath)
        self._basepath = basepath
        self._contents = l
        self._index = 0
        super().__init__()

    def __iter__(self):
        return self

    def __next__(self):
        """ d
        Iterate through whatever remote file tree is involved and return individual file objects
        :return: StorageObject representing a single file object
        :rtype: StorageObject
        """
        i = self._index
        c = self._contents
        if i < len(c):
            self._index = i + 1
            return LocalStorageItem(c[i])
        else:
            raise StopIteration

    def __str__(self):
        return self._basepath




class LocalStorage(Storage):
    """
    Represents a "connection" to remote storage. All methods may implement additional OPTIONAL parameters, but generally
    should store within itself any other information needed to execute searches, puts, gets, and deletes.
    The exception being the __init__ method, which may involve more strict and platform-specific requirements, such as
    credentials.
    ...
    Attributes
    ----------
    ItemTypes
    """
    ItemTypes = [StorageItem]
    stream_support = False

    def __init__(self):
        super().__init__()
        self.ItemTypes = [StorageItem]


    def search(self, basepath: str = ""):
        """
        Return an interator representing all objects within a specified search of the remote storage.
        Should generally return objects that are children of, or begin with basepath.
        :param basepath: Allow limiting to sub "directories" or other organizational schemas
        :return: StorageObjectIter representing all files within basepath
        :rtype: StorageObjectIter
        """
        return LocalStorageSearchIter(basepath)


    def put(self, o: LocalStorageItem, filepath: str, prefix: str = ""):
        """
        Upload file from filepath to remote storage
        :param o: Remote storage representation
        :param filepath: local file location
        :return: Success?
        :rtype: bool
        """
        try:
            os.makedirs(Path(prefix + o.path).parent, exist_ok=True)
            shutil.copy(filepath, prefix + o.path)
        except SameFileError:
            pass
        return True

    def put_stream(self, o: StorageItem, file: io.IOBase, prefix: str = ""):
        """
        Download file from remote storage to local IO Stream
        :param o: Remote storage object representation
        :param file: IO object to write a binary stream to
        :return: Success?
        :rtype: bool
        """
        raise NotImplemented


    def get(self, o: LocalStorageItem, filepath: str = ""):
        """
        Download file from remote storage to filepath
        :param o: Remote storage object representation
        :param filepath: optional information to associate with the remote object, eg s3 tags, expiry, or storage settings
        :return: Success?
        :rtype: bool
        """
        try:
            os.makedirs(Path(filepath).parent, exist_ok=True)
            shutil.copy(o.path, filepath)
        except SameFileError:
            pass
        return True


    def fetch(self, o: LocalStorageItem):
        """
        Fetch details about a remote object and return a object containing that information
        The intent is for a user to create a base StorageItem object containing enough information to locate the file
            in question (say, a remote path for example), and then use this method to fetch all actual metadata about
            the object that was not known prior. The method should not modify o, but should create a new StorageItem
        :param o: Remote storage object representation
        :return: o?
        :rtype: StorageObject
        """
        return LocalStorageItem.from_file(o.path)

    def get_stream(self, o: StorageItem, file: io.IOBase):
        """
        Download file from remote storage to local IO Stream
        :param o: Remote storage object representation
        :param file: IO object to write a binary stream to
        :return: Success?
        :rtype: bool
        """
        raise NotImplemented


    def delete(self, o: LocalStorageItem):
        """
        Delete file from remote storage
        :param o: Remote storage representation
        :return: Success?
        :rtype: bool
        """
        os.remove(o.path)


from ..core.storage import StorageItem, StorageSearchIter, Storage
from ..core.exceptions import ObjectNotFoundError, UnsupportedItemTypeError
import boto3
import botocore
import os
import io
import builtins
import pathlib


class S3StorageItem(StorageItem):
    pass

class S3StorageSearchIter(StorageSearchIter):
    pass

class S3Storage(Storage):
    pass

class S3StorageItem(StorageItem):
    DeleteMarker = None
    AcceptRanges = None
    Expiration = None
    Restore = None
    ArchiveStatus = None
    LastModified = None
    ContentLength = None
    ETag = None
    MissingMeta = None
    VersionId = None
    CacheControl = None
    ContentDisposition = None
    ContentEncoding = None
    ContentLanguage = None
    ContentType = None
    Expires = None
    WebsiteRedirectLocation = None
    ServerSideEncryption = None
    Metadata = None
    SSECustomerAlgorithm = None
    SSECustomerKeyMD5 = None
    SSEKMSKeyId = None
    BucketKeyEnabled = None
    StorageClass = None
    RequestCharged = None
    ReplicationStatus = None
    PartsCount = None
    ObjectLockMode = None
    ObjectLockRetainUntilDate = None
    ObjectLockLegalHoldStatus = None
    Key = None

    @staticmethod
    def from_file(path, remote_path="", name=None):
        if name is None:
            name = pathlib.PurePath(path).name
        return S3StorageItem(Key=name)

    @staticmethod
    def from_storage(s3: S3Storage, key: str):
        return s3.fetch(S3StorageItem(Key=key))

    @staticmethod
    def from_dict(d: dict, Key=None):
        if Key is None:
            if 'Key' not in d:
                raise ValueError("Must specify a key")
            else:
                Key = d['Key']
        i = S3StorageItem(Key=Key)
        for key in d:
            if key in i.__dict__:
                i.__dict__[key] = d[key]
        i.Key = Key # Enforces manually specified Key if there was also a Key in the dict
        return i

    def __init__(self, Key=None):
        self.Key = Key
        if self.Key is None:
            raise ValueError("Must specify a key")
        super().__init__()

    @property
    def name(self):
        return self.Key.split('/')[-1]

    @property
    def size(self):
        return self.ContentLength

    @property
    def remote_path(self):
        return self.Key

    def __str__(self):
        return self.Key


class S3StorageSearchIter(StorageSearchIter):
    def __init__(self,
                 client: [boto3.Session, botocore.client] = None,
                 bucket: str = None, basepath: str = "",
                 pagesize: int = 100):
        self._client = client
        self._bucket = bucket
        self._basepath = basepath
        self._pagesize = pagesize

        self._response = self._client.list_objects_v2(
            Bucket=self._bucket,
            MaxKeys=self._pagesize,
            Prefix=self._basepath
        )
        self._index = 0
        super().__init__()

    def __iter__(self):
        return self

    def _fetchNewS3Page(self):
        if not self._response['IsTruncated']:
            return False
        self._response = self._client.list_objects_v2(
            Bucket=self._bucket,
            MaxKeys=self._pagesize,
            Prefix=self._basepath,
            ContinuationToken=self._response['NextContinuationToken']
        )
        return True

    def __str__(self):
        return f"s3://{self._bucket}/{self._basepath}*"

    def __next__(self):
        index = self._index
        if index >= self._pagesize:
            if not self._fetchNewS3Page():
                raise StopIteration
            index = 0
        content = self._response['Contents']
        if index >= len(content):
            raise StopIteration
        ret = content[index]
        self._index = index + 1
        return S3StorageItem.from_dict(d=ret)


class S3Storage(Storage):
    _client = None
    _bucket = None
    ItemTypes = [S3StorageItem]
    stream_support = False

    def __init__(self,
                 session: [boto3.Session, botocore.client] = None,
                 bucket_name: str = None,
                 aws_access_key_id = None,
                 aws_secret_access_key = None,
                 endpoint_url = None):
        if bucket_name is None:
            raise ValueError("Must specify a bucket")
        if session is None:
            if (aws_access_key_id is None) and (aws_secret_access_key is None):
                self._client = boto3.Session().client('s3')
            if (aws_access_key_id is None) != (aws_secret_access_key is None):
                raise ValueError("Incomplete credentials")
            else:
                self._client = boto3.Session().client(
                    service_name='s3',
                    aws_access_key_id=aws_access_key_id,
                    aws_secret_access_key=aws_secret_access_key,
                    endpoint_url=endpoint_url
                )
        elif isinstance(session, boto3.Session):
            self._client = session.client(
                service_name='s3',
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                endpoint_url=endpoint_url
            )
        else: # we're gonna assume this is a client now, but we can't check easily because boto3 is built really weirdly
            self._client = session
        self._bucket = bucket_name
        super().__init__()

    def search(self, basepath: str = "", pagesize: int = 100):
        return S3StorageSearchIter(self._client, self._bucket, basepath=basepath, pagesize=pagesize)

    def put(self, o: S3StorageItem, filepath: str, prefix: str = ""):
        if type(o) not in self.ItemTypes:
            raise UnsupportedItemTypeError

        self._client.upload_file(
            filepath,
            self._bucket,
            prefix + o.Key
        )

        return True

    def put_stream(self, o: S3StorageItem, file: io.IOBase, callback = None, config = None, prefix: str = ""):
        """
        Download file from remote storage to local IO Stream
        :param o: Remote storage object representation
        :param file: IO object to write a binary stream to
        :param callback: Callback for S3 download accepting number of bytes
        :param config: Boto3 Transfer Config for controlling bytes
        :param prefix: Prefix to append to key
        :return: Success?
        :rtype: bool
        """
        self._client.upload_fileobj(
            Fileobj=file,
            Bucket=self._bucket,
            Key=prefix + o.Key,
            Callback=callback,
            Config=config
        )
        raise NotImplemented

    def get(self, o: S3StorageItem, filepath=""):
        if type(o) not in self.ItemTypes:
            raise ValueError(f"Unsupported Item type {str(type(o))}")
        key = o.Key
        tmp = filepath.rsplit('/', 1)
        if filepath == "":
            filepath = tmp[1]
        if tmp != "":
            os.makedirs(tmp[0], exist_ok=True)
        self._client.download_file(self._bucket, key, filepath)
        return True

    def fetch(self, o: S3StorageItem):
        """
        Fetch object info
        :param o: Remote storage object representation
        :return: Duplicate o containing additional metadata
        :rtype: S3StorageItem
        """
        try:
            ret = self._client.head_object(
                Bucket=self._bucket,
                Key=o.Key,
            )
            return S3StorageItem.from_dict(ret, Key=o.Key)
        except self._client.exceptions.NoSuchKey as e:
            raise ObjectNotFoundError(f"Key f{o.Key} does not exist")

    def get_stream(self, o: S3StorageItem, file: io.IOBase, callback = None, config = None):
        """
        Download file from remote storage to local IO Stream
        :param o: Remote storage object representation
        :param file: IO object to write a binary stream to
        :param callback: Callback for S3 download accepting number of bytes
        :param config: Boto3 Transfer Config for controlling bytes
        :return: Success?
        :rtype: bool
        """
        self._client.download_fileobj(
            Fileobj=file,
            Bucket=self._bucket,
            Key=o.Key,
            Callback=callback,
            Config=config
        )
        return True

    def delete(self, o: StorageItem):
        if type(o) not in self.ItemTypes:
            raise ValueError(f"Unsupported Item type {str(type(o))}")
        key = o.Key

        self._client.delete_object(
            Bucket=self._bucket,
            Key=key
        )
        return True

    def __str__(self):
        return f"s3://{self._bucket}"

